"""List all installed seeds."""
from enum import Enum
from pydoc import locate
from types import ModuleType
from datoso import __app_name__

class PluginType(Enum):
    SEED = 'seed'
    PLUGIN = 'plugin'

def get(plugin, module, plugin_type):
    if module:
        return locate(f'{__app_name__}_{plugin_type}_{plugin}.{module}')
    return locate(f'{__app_name__}_{plugin_type}_{plugin}')

def installed(plugin_type):
    import importlib
    import pkgutil
    return {
        name: importlib.import_module(name)
        for finder, name, ispkg
        in pkgutil.iter_modules()
        if name.startswith(f'{__app_name__}_{plugin_type}_')
    }

def description(plugin, plugin_type):
    plugin = plugin if isinstance(plugin, ModuleType) else locate(f'{__app_name__}_{plugin_type}_{plugin}')
    return plugin.__description__

def get_seed(seed, module=None):
    return get(seed, module, PluginType.SEED.value)

def installed_seeds():
    return installed(PluginType.SEED.value)

def seed_description(seed):
    return description(seed, PluginType.SEED.value)

def get_plugin(plugin, module=None):
    return get(plugin, module, PluginType.PLUGIN.value)

def installed_plugins():
    return installed(PluginType.PLUGIN.value)

def plugin_description(plugin):
    return description(plugin, PluginType.PLUGIN.value)



"""
    Database models for the datfile.
"""
# pylint: disable=too-few-public-methods
from typing import Optional

from pydantic import BaseModel, Extra
from tinydb import Query
from datoso.database import DB

class Database:
    """ TinyDB wrapper (). """
    DB = DB
    table = None

class DatabaseModel(BaseModel): # BaseModel needed for initialization, TODO: find a better way
    """ Base model for TinyDB. """
    class Config:
        """ Pydantic default config. """
        extra = Extra.allow
        arbitrary_types_allowed = True

    _DB = Database()
    _id: int = None
    _table_name: str = None

    def __init__(self, **kwargs) -> None:
        self._DB.DB = DB  # pylint: disable=invalid-name
        self._DB.table = DB.table(self._table_name)
        super().__init__(**kwargs)

    def load(self):
        """ Load record from the database. """
        result = self._DB.table.search(self.query())
        if result:
            self.__dict__.update(result[0])

    def save(self):
        """ Save record to the database. """
        query = Query()
        if self._id:
            self._DB.table.upsert(self.dict(), query.id == self._id)
        else:
            self._id = self._DB.table.upsert(self.dict(), self.query())

    def query(self):
        """ Query to update or load a record. """
        query = Query()
        return query.id == self._id

    def close(self):
        """ Close the database. """
        self._DB.table.storage.flush()



class Dat(DatabaseModel):
    """ Dat file model. """
    _table_name = 'dats'
    name: str
    modifier: Optional[str]
    company: Optional[str]
    system: Optional[str]
    seed: str

    def query(self):
        """ Query to update or load a record. """
        query = Query()
        return (query.name == self.name) & (query.seed == self.seed)

    def is_enabled(self):
        """ Check if the dat is enabled. """
        return not getattr(self, 'status', None) or self.status == 'enabled' # pylint: disable=no-member


class Seed(DatabaseModel):
    """ Repo file model. """
    _table_name = 'repos'
    name: str

    def query(self):
        """ Query to update or load a record. """
        query = Query()
        return query.name == self.name

class System(DatabaseModel):
    """ System file model. """
    _table_name = 'systems'
    company: Optional[str]
    system: str

    def query(self):
        """ Query to update or load a record. """
        query = Query()
        return (query.company == self.company) & (query.system == self.system)

    @staticmethod
    def all():
        """ Get all systems. """
        System._DB.DB = DB  # pylint: disable=invalid-name
        System._DB.table = DB.table(System._table_name)
        return System._DB.table.all()


""" Logging configuration for datoso """
import logging
import os

from datoso.helpers import Bcolors, FileUtils
from .configuration import config

log_level = config['LOG'].get('LogLevel', logging.DEBUG)
formatter = logging.Formatter("[%(asctime)s - %(levelname)s] - %(message)s", "%Y-%m-%d %H:%M:%S")

class TrimmedFileHandler(logging.FileHandler):
    """ File handler that removes color codes from the log message. """

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        # Get the log message
        msg = self.format(record).strip()

        # Write the trimmed message to the file
        if msg:
            try:
                self.stream.write(Bcolors.remove_color(msg) + self.terminator)
                self.flush()
            except Exception:
                self.handleError(record)



class TrimmedStreamHandler(logging.StreamHandler):
    """ Stream handler that allows to stdout while logging. """

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        try:
            msg = self.format(record)
            stream = self.stream
            # issue 35046: merged two stream.writes into one.
            stream.write(msg)
            self.flush()
        except RecursionError:  # See issue 36272
            raise
        except Exception:
            self.handleError(record)

# Get root logger
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

def enable_logging():
    """ Enable logging to file """
    log_path = FileUtils.parse_folder(config.get('PATHS','DatosoPath', fallback='~/.datoso'))
    os.makedirs(log_path, exist_ok=True)
    log_file = os.path.join(log_path, config['LOG'].get('LogFile', 'datoso.log'))
    file_handler = TrimmedFileHandler(log_file)
    file_handler.setLevel(log_level)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

# Create a file handler
if config.getboolean('LOG', 'Logging', fallback=False):
    enable_logging()

# Create a stream handler
stream_handler = TrimmedStreamHandler()
stream_handler.setLevel(logging.INFO)
if config.getboolean('COMMAND', 'Verbose', fallback=False):
    stream_handler.setLevel(logging.DEBUG)
if config.getboolean('COMMAND', 'Quiet', fallback=False):
    stream_handler.setLevel(logging.WARNING)
logger.addHandler(stream_handler)

def set_verbosity(verbosity):
    """ Set the log level of the stream handler """
    stream_handler.setLevel(verbosity)

def set_quiet():
    """ Set the log level of the stream handler to warning """
    stream_handler.setLevel(logging.WARNING)

def set_verbose():
    """ Set the log level of the stream handler to debug """
    stream_handler.setLevel(logging.DEBUG)

def get_verbosity():
    """ Get the log level of the stream handler """
    return stream_handler.level

def get_file_level():
    """ Get the log level of the file handler """
    for handler in logger.handlers:
        if isinstance(handler, TrimmedFileHandler):
            return handler.level
    return None

# for handler in logger.handlers:
#     print(get_file_level())
#     print(get_stream_level())




import os
from datoso.helpers import FileUtils
from datoso.configuration import config


class Folders:
    base: str = None
    download: str = None
    dats: str = None
    backup: str = None
    extras: dict = None

    def __init__(self, seed, extras=[]) -> None:
        self.base = FileUtils.parse_folder(config['PATHS'].get('DownloadPath'))
        self.download = os.path.join(self.base, seed)
        self.dats = os.path.join(self.download, 'dats')
        self.backup = os.path.join(self.download, 'backup')
        self.extras = []
        for extra in extras:
            self.extras.append(os.path.join(self.dats, extra))

    def clean_dats(self):
        FileUtils.remove(self.dats)
        os.makedirs(self.dats, exist_ok=True)

    def create_all(self):
        os.makedirs(self.download, exist_ok=True)
        os.makedirs(self.dats, exist_ok=True)
        os.makedirs(self.backup, exist_ok=True)
        for extra in self.extras:
            os.makedirs(extra, exist_ok=True)




"""
    Shortcuts for models.
"""
__all__ = ['DatabaseModel', 'Dat', 'Seed', 'System']

from .datfile import DatabaseModel
from .datfile import Dat
from .datfile import Seed
from .datfile import System



""" Rules class. """
from pydoc import locate
from datoso.helpers.plugins import installed_seeds

class Rules:
    """ Rules class. """
    _rules = []

    def __init__(self):
        for seed, _ in installed_seeds().items():
            rules = locate(f'{seed}.rules')
            self._rules.extend(rules.get_rules())
        self._rules.sort(key=lambda x: x['priority'], reverse=True)

    @property
    def rules(self):
        """ Return the rules. """
        return self._rules



""" Fetch and Process Commands for Seeds """
import os
import re
from datoso.helpers.plugins import get_seed
from datoso.helpers import Bcolors, FileUtils
from datoso.configuration import config
from datoso.helpers.executor import Command
from datoso.actions.processor import Processor

class Seed:
    """ Seed class """
    name = None
    path = None
    actions = {}
    # working_path = os.path.abspath(os.path.join(os.getcwd(), config.get('PATHS', 'WorkingPath')))
    status_to_show = ['Updated', 'Created', 'Error', 'Disabled', 'Deduped', 'No Action Taken, Newer Found']

    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)
        actions = get_seed(self.name, 'actions')
        if actions:
            self.actions = actions.get_actions()


    def fetch(self):
        """ Fetch seed """
        fetch = get_seed(self.name, 'fetch')
        fetch.fetch()


    def format_actions(self, actions, data: dict = {}):
        """ Format actions. """
        for action in actions:
            for action_name, action_value in action.items():
                if isinstance(action_value, str):
                    action[action_name] = action_value.format(**data)
        return actions


    def process_dats(self, fltr=None, actions_to_execute=None):
        """ Process dats."""
        def delete_line(line):
            # pylint: disable=expression-not-assigned
            [print('\b \b', end='') for x in range(0, len(line))]
            print(' ' * (len(line)), end='')
            print('\r', end='')
        def get_preffix(name) -> str:
            seed = get_seed(name)
            return seed.__preffix__ if seed else name
        tmp_path = config['PATHS'].get('DownloadPath', 'tmp')
        dat_origin = os.path.join(FileUtils.parse_folder(tmp_path), get_preffix(self.name), 'dats')
        line = ''
        for path, actions in self.actions.items():
            new_path = path.format(dat_origin=dat_origin)
            actions = self.format_actions(actions, data={'dat_destination': config['PATHS'].get('DatPath', 'DatRoot')})
            if actions_to_execute:
                actions = [x for x in actions if x['action'] in actions_to_execute]
            for file in os.listdir(new_path) if os.path.isdir(new_path) else []:
                if config['PROCESS'].get('DatIgnoreRegEx'):
                    ignore_regex = re.compile(config['PROCESS']['DatIgnoreRegEx'])
                    if ignore_regex.match(file):
                        continue

                if (not file.endswith(('.dat', '.xml')) \
                    and not os.path.isdir(os.path.join(new_path,file))) \
                    or (fltr and fltr not in file):
                    continue

                if not config.getboolean('COMMAND', 'Quiet', fallback=False):
                    delete_line(line)
                    line = f'Processing {Bcolors.OKCYAN}{file}{Bcolors.ENDC}'
                    print(line, end=' ', flush=True)
                procesor = Processor(seed=self.name, file=f'{new_path}/{file}', actions=actions)
                # output = [x for x in procesor.process() if (x in self.status_to_show or Command.verbose)]
                output = []
                for process in procesor.process():
                    if process in self.status_to_show or Command.verbose:
                        output.append(process)
                    if process == 'Error':
                        break
                if 'Deleted' in output and 'Ignored' in output:
                    output.append('Disabled')
                if not config.getboolean('COMMAND', 'Quiet', fallback=False):
                    # [print('\b \b', end='') for x in range(0, len(line))]
                    delete_line(line)
                    line = f'Processed {Bcolors.OKCYAN}{file}{Bcolors.ENDC}'
                    print(line, end=' ', flush=True)

                if output and not config.getboolean('COMMAND', 'Quiet', fallback=False):
                    line += str(output)+' '
                    print(output, end=' ', flush=True)
                if output or config.getboolean('COMMAND', 'Verbose', fallback=False):
                    line = ''
                    print(line)
        delete_line(line)



""" Configuration module. """
import configparser
import os

from datoso import __app_name__, ROOT_FOLDER

config = configparser.ConfigParser(allow_no_value=True)
config.optionxform = lambda option: option
config.read(os.path.join(ROOT_FOLDER, 'datoso.ini'))
config.read(os.path.expanduser('~/.datosorc'))
config.read(os.path.join(os.getcwd(), '.datosorc'))


def get_seed_name(seed):
    """ Get seed name. """
    return seed.replace(f'{__app_name__}_seed_', '')



"""
    Database module
"""
import os

from typing import Dict, Any
from tinydb import TinyDB, JSONStorage
from tinydb.middlewares import CachingMiddleware

from datoso.helpers import FileUtils
from datoso.configuration import config

database_path = FileUtils.parse_folder(config['PATHS'].get('DatosoPath','~/.datoso'))
os.makedirs(f"{database_path}", exist_ok=True)

DATABASE_URL = os.path.join(database_path, config['PATHS'].get('DatabaseFile','datoso.json'))

class JSONStorageWithBackup(JSONStorage):
    """ TinyDB JSON storage with backup. """
    path: str = DATABASE_URL

    def __init__(self, path: str, create_dirs=False, encoding=None, access_mode='r+', **kwargs):
        self.path = path
        super().__init__(path, create_dirs, encoding, access_mode, **kwargs)

    def write(self, data: Dict[str, Dict[str, Any]]):
        """ Write data to the storage. """
        self.make_backup()
        super().write(data)

    def make_backup(self):
        """ Make a backup of the database. """
        FileUtils.copy(self.path, f"{self.path}.bak")


DB = TinyDB(DATABASE_URL, storage=CachingMiddleware(JSONStorageWithBackup), indent=4)

# DB = TinyDB(DATABASE_URL, storage=JSONStorageWithBackup, indent=4)



""" Command executor

This module provides a wrapper for subprocess.Popen to execute commands.
"""
import logging
from subprocess import PIPE, CalledProcessError, Popen
from datoso.configuration import config

class Command:
    """ Subprocess wrapper """
    quiet = config.getboolean('COMMAND', 'Quiet', fallback=False)
    verbose = config.getboolean('COMMAND', 'Verbose', fallback=False)
    logging = config.getboolean('LOG', 'Logging', fallback=False)
    logfile = config.get('LOG', 'LogFile', fallback='datoso.log')

    def __init__(self, command):
        self.command = command

    @staticmethod
    def execute(args, cwd=None, env=None, universal_newlines=True, shell=False):
        """ Execute a command, allows to write to log and stdout. """
        # pylint: disable=consider-using-with
        def execute(args):
            popen = Popen(args, stdout=PIPE, universal_newlines=universal_newlines, cwd=cwd, env=env, shell=shell)
            for stdout_line in iter(popen.stdout.readline, ""):
                yield stdout_line
            popen.stdout.close()
            return_code = popen.wait()
            if return_code:
                raise CalledProcessError(return_code, args)

        for output in execute(args):
            logging.info(output)
